// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.

 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <time.h>
#include <stdlib.h>

#include "role_offensive_half.h"
#include <rcsc/action/body_smart_kick.h>

#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"

#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/player/audio_sensor.h>
#include <rcsc/player/freeform_parser.h>
#include <rcsc/player/abstract_player_object.h>

#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>


#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_memory.h>
#include <rcsc/common/say_message_parser.h>
#include <rcsc/common/free_message_parser.h>

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doKick( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();	
  srand((unsigned) time(NULL));
  const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
  const rcsc::PlayerObject * nearest_opp
    = ( opps.empty()
	? static_cast< rcsc::PlayerObject * >( 0 )
	: opps.front() );
	  
  rcsc::AngleDeg opp_angle = nearest_opp->angleFromSelf();
  const rcsc::PlayerObject * goalie = agent->world().getOpponentGoalie();
  rcsc::AngleDeg goal_angle = goalie->angleFromSelf();
  

    switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:

      {
	if (wm.offsideLineX() - wm.self().pos().x < 1.0){
	  Bhv_BasicOffensiveKick().execute( agent );
	  break;
	}
	const double nearest_opp_dist = ( nearest_opp
					  ? nearest_opp->distFromSelf()
					  : 1000.0 );
	if (wm.offsideLineX() - wm.self().pos().x < 5.0 && nearest_opp_dist < 3.0){
	  srand((unsigned) time(NULL));
	  int random = rand();
	  double passrand = random % 200;	//make random num
	  passrand -= 100;
	  double passy = 33 * passrand / 100;  //make through_point by random
	  double passx = std::min( wm.self().pos().x + 15, 45.0);	  
	  rcsc::Vector2D through_point(passx , passy); //target_point

	  double first_speed = rcsc::calc_first_term_geom_series_last(
								      0.5,
								      agent->world().ball().pos().dist(through_point),
								      rcsc::ServerParam::i().ballDecay()
								      ); //calc first pass speed
	  first_speed = std::min(first_speed, rcsc::ServerParam::i().ballSpeedMax());
	  
	  if (rcsc::Body_SmartKick(through_point, first_speed, first_speed * 0.8, 2).execute(agent)){
	    std::cout << "kick_off_mid" << passy<< std::endl;
	    rcsc::Vector2D ball_vel (passx - wm.self().pos().x, passy - wm.self().pos().y);
	    agent->addSayMessage(new rcsc::PassMessage(
						       11,
						       through_point,
						       wm.self().pos(),
						       ball_vel
						       ));
	    rcsc::dlog.addText (rcsc::Logger::SENSOR, __FILE__": Said X:%f Y:%f", 
				through_point.x, through_point.y);
	    break;
	  }
	}
      }
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleOffensiveHalf::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    }
}
