// -*-c++-*-

/*
 *Copyright:

 Copyright (C) Hidehisa AKIYAMA

 This code is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2, or (at your option)
 any later version.srand((unsigned) time(NULL));


 This code is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this code; see the file COPYING.  If not, write to
 the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

 *EndCopyright:
 */

/////////////////////////////////////////////////////////////////////

#ifdef HAVE_CONFIG_H
#include <config.h>

//#include <iostream.h>
#endif
#include <stdlib.h>
#include <time.h>

#include "role_side_forward.h"

#include <rcsc/action/body_smart_kick.h>
//#include "body_smart_kick.h"
//#include "kick_table.h"
#include <rcsc/action/bhv_shoot2008.h>
#include "strategy.h"

#include "bhv_basic_offensive_kick.h"
#include "bhv_basic_move.h"

#include <rcsc/math_util.h>
#include <rcsc/formation/formation.h>

#include <rcsc/player/player_agent.h>
#include <rcsc/player/intercept_table.h>
#include <rcsc/player/debug_client.h>
#include <rcsc/player/say_message_builder.h>
#include <rcsc/player/audio_sensor.h>
#include <rcsc/player/freeform_parser.h>
#include <rcsc/player/abstract_player_object.h>
#include <rcsc/action/body_kick_multi_step.h>
#include <rcsc/action/neck_turn_to_goalie_or_scan.h>


#include <rcsc/common/logger.h>
#include <rcsc/common/server_param.h>
#include <rcsc/common/audio_memory.h>
#include <rcsc/common/say_message_parser.h>
#include <rcsc/common/free_message_parser.h>

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::execute( rcsc::PlayerAgent * agent )
{
    bool kickable = agent->world().self().isKickable();
    if ( agent->world().existKickableTeammate()
         && agent->world().teammatesFromBall().front()->distFromBall()
         < agent->world().ball().distFromSelf() )
    {
        kickable = false;
    }

    if ( kickable )
    {
        doKick( agent );
    }
    else
    {
        doMove( agent );
    }

}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doKick( rcsc::PlayerAgent * agent )
{
  const rcsc::WorldModel & wm = agent->world();	
  srand((unsigned) time(NULL));
  const rcsc::PlayerPtrCont & opps = wm.opponentsFromSelf();
  const rcsc::PlayerObject * nearest_opp
    = ( opps.empty()
	? static_cast< rcsc::PlayerObject * >( 0 )
	: opps.front() );
	  
  rcsc::AngleDeg opp_angle = nearest_opp->angleFromSelf();
  const rcsc::PlayerObject * goalie = agent->world().getOpponentGoalie();
  rcsc::AngleDeg goal_angle = goalie->angleFromSelf();
  

  switch ( Strategy::get_ball_area( agent->world().ball().pos() ) ) {
  case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    case Strategy::BA_OffMidField:
    case Strategy::BA_DribbleAttack:

      {
	Bhv_BasicOffensiveKick().execute( agent );
	if (wm.offsideLineX() - wm.self().pos().x < 1.0){
	  break;
	}
	srand((unsigned) time(NULL));
	int random = rand();
	//rcsc::dlog.addText (rcsc::Logger::SENSOR, __FILE__":opp:%f goal:%f",opp_angle.degree(), goal_angle.degree());
	//if access to goal_angle or opp_angle, player will be dead
	//pass to center forward

	if (wm.offsideLineX() - wm.self().pos().x < 7.0){	//near the off-side line
	  std::cout << "pass?" << std::endl;
	  double passy;
	  double passrand = random % 100;	//make random num
	  if (wm.self().pos().y > 0){
	    passy = 15 * passrand / 100;  //make through_point by random
	  }else{
	    passy = -15 * passrand / 100;  //make through_point by random
	  }
	  //double passy = nearest_opp->PlayerObject().pos().y;
	  double passx = std::min( wm.self().pos().x + 15, 45.0);
	  
	  rcsc::Vector2D through_point(passx , passy); //target_point

	  //rcsc::AngleDeg target_angle = through_point->angleFromSelf();
	  
	 
	  //std::cout << opp_angle.degree() - goal_angle.degree()  << std::endl;
	  
	  if (true){
	    double first_speed = rcsc::calc_first_term_geom_series_last(
									1.0,
									agent->world().ball().pos().dist(through_point),
									rcsc::ServerParam::i().ballDecay()
									); //calc first pass speed
	    first_speed = std::min(first_speed, rcsc::ServerParam::i().ballSpeedMax());
	    //get nearest opp dist
	    
	    const double nearest_opp_dist = ( nearest_opp
					      ? nearest_opp->distFromSelf()
					      : 1000.0 );
	    //if opp is far
	    if (nearest_opp_dist > 3.0){
	      if (rcsc::Body_SmartKick(through_point, first_speed, first_speed * 0.8, 2).execute(agent)){
		std::cout << "kick?" << passy<< std::endl;
		rcsc::Vector2D ball_vel (passx - wm.self().pos().x, passy - wm.self().pos().y);
		agent->addSayMessage(new rcsc::PassMessage(
							   11,
							   through_point,
							   wm.self().pos(),
							   ball_vel
							   ));
		rcsc::dlog.addText (rcsc::Logger::SENSOR, __FILE__": Said X:%f Y:%f", 
				    through_point.x, through_point.y);
	      }
	    }else{
	      Bhv_BasicOffensiveKick().execute(agent);
	    }
	  }
	}
	break;
      }
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
	if (rcsc::Bhv_Shoot2008().execute( agent )){
		break;
	}
    default:
        Bhv_BasicOffensiveKick().execute( agent );
        break;
    }
}

/*-------------------------------------------------------------------*/
/*!

*/
void
RoleSideForward::doMove( rcsc::PlayerAgent * agent )
{
    const rcsc::WorldModel & wm = agent->world();

    int ball_reach_step = 0;
    if ( ! wm.existKickableTeammate()
         && ! wm.existKickableOpponent() )
    {
        ball_reach_step
            = std::min( wm.interceptTable()->teammateReachCycle(),
                        wm.interceptTable()->opponentReachCycle() );
    }
    const rcsc::Vector2D base_pos
        = wm.ball().inertiaPoint( ball_reach_step );

    rcsc::Vector2D home_pos
        = formation().getPosition( agent->config().playerNumber(),
                                   base_pos );
    if ( rcsc::ServerParam::i().useOffside() )
    {
        home_pos.x = std::min( home_pos.x, wm.offsideLineX() - 1.0 );
    }

    rcsc::dlog.addText( rcsc::Logger::ROLE,
                        "%s: HOME POSITION=(%.2f, %.2f) base_point(%.1f %.1f)"
                        ,__FILE__,
                        home_pos.x, home_pos.y,
                        base_pos.x, base_pos.y );

    switch ( Strategy::get_ball_area( base_pos ) ) {
    case Strategy::BA_CrossBlock:
    case Strategy::BA_Stopper:
    case Strategy::BA_Danger:
    case Strategy::BA_DribbleBlock:
    case Strategy::BA_DefMidField:
    case Strategy::BA_DribbleAttack:
    case Strategy::BA_OffMidField:
    case Strategy::BA_Cross:
    case Strategy::BA_ShootChance:
    default:
        Bhv_BasicMove( home_pos ).execute( agent );
        break;
    }
}
